#!/usr/bin/env python
# synchronise.py - mercurial extension
# Copyright (C) 2010 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


'tools for synchronising a one-man repository'

from mercurial import util, hg, cmdutil, url, commands
from mercurial.node import nullrev
from mercurial.i18n import _
from mercurial.__version__ import version as hgVersion

def getTaskFilePath(repo):
    return repo.join('synchronise_task')

def setCurrentTask(repo, task):
    open(getTaskFilePath(repo), 'w').write(task)

def getCurrentTask(repo):
    try:
        return open(getTaskFilePath(repo), 'r').read()
    except IOError:
        return ''

WORKON_OPTIONS = []
def doWorkon(ui, repo, task=None):
    '''
    Specifies the task that is currently being worked on in this repository.
    Calls to hg synchronise will use this task in the commit message. When the
    task is complete, use hg complete.

    If executed with no argument, will display the current task.
    '''
    if not repo.local():
        raise util.Abort(_('the repository is not local'))
    if task is None:
        task = getCurrentTask(repo)
        if task:
            ui.note(_('Current task:\n'))
            ui.write('%s\n' % (task,))
        else:
            ui.warn(_('there is no current task\n'))
    else:
        setCurrentTask(repo, task)
        ui.status(_('current task set to %r\n') % (task,))

UPSYNC_OPTIONS = []
def doUpsync(ui, repo, dest=None):
    '''
    If there are no uncommitted changes, takes no action, otherwise behaves the
    same as hg synchronise.
    '''
    message = _('Auto save')
    task = getCurrentTask(repo)
    if task:
        message += ': %s' % (task,)

    node = repo.commit(message)
    if node is None:
        ui.status(_('nothing changed\n'))
        return
    elif createsNewHead(repo, node):
        ui.status(_('created new head\n'))
    else:
        ui.status(_('saved changes\n'))

    return mainSynchronise(ui, repo, dest)

SYNCHRONISE_OPTIONS = []
def doSynchronise(ui, repo, dest=None):
    '''
    Synchronises the current repository with the server, giving an error message
    if any merges are required. If there are uncommitted changes, they will be
    committed using a commit message containing the task set by hg workon. If no
    task has been specified using hg workon since the last call to hg complete,
    a default commit message will be used, and a warning will be displayed.
    '''
    message = _('Auto save')
    task = getCurrentTask(repo)
    if task:
        message += ': %s' % (task,)

    node = repo.commit(message)
    if node is None:
        pass    # No changes.
    elif createsNewHead(repo, node):
        ui.status(_('created new head\n'))
    else:
        ui.status(_('saved changes\n'))

    return mainSynchronise(ui, repo, dest)

def mainSynchronise(ui, repo, dest=None):
    if hgVersion >= '1.6.3':
        dest, branches = hg.parseurl(ui.expandpath(dest or 'default-push',
                dest or 'default'))
        revs, checkout = hg.addbranchrevs(repo, repo, branches, None)
        other = hg.repository(hg.remoteui(repo, {}), dest)
    else:
        dest, revs, checkout = hg.parseurl(ui.expandpath(dest or 'default-push',
                dest or 'default'))
        other = hg.repository(cmdutil.remoteui(repo, {}), dest)

    ui.status(_('pushing to %s\n') % url.hidepassword(dest))
    pushRepo(ui, repo, other)
    ui.status(_('pulling from %s\n') % url.hidepassword(dest))
    pullRepo(ui, repo, other, checkout)

def pushRepo(ui, repo, other):
    '''
    Copied from commands.push().
    '''
    # push subrepos depth-first for coherent ordering
    c = repo['']
    subs = c.substate # only repos that are committed
    for s in sorted(subs):
        c.sub(s).push(force=True)

    r = repo.push(other, force=True)
    return r == 0

def pullRepo(ui, repo, other, checkout):
    modheads = repo.pull(other, force=True)
    if checkout:
        checkout = str(repo.changelog.rev(other.lookup(checkout)))
    return commands.postincoming(ui, repo, modheads, True, checkout)
 

COMPLETE_OPTIONS = []
def doComplete(ui, repo):
    '''
    Indicates that the current task (as specified to hg workon) is complete.
    This will clear the current task message used by hg synchronise, commit the
    changes with a suitable log message, and synchronise with the server.
    '''
    if not repo.local():
        raise util.Abort(_('the repository is not local'))

    task = getCurrentTask(repo)
    if not task:
        raise util.Abort(_('no active task to complete'))

    node = repo.commit(_('COMPLETED: %s') % (task,), force=True)
    if createsNewHead(repo, node):
        ui.status(_('created new head\n'))

    setCurrentTask(repo, '')

    mainSynchronise(ui, repo)

def createsNewHead(repo, node):
    cl = repo.changelog
    rev = cl.rev(node)
    parents = cl.parentrevs(rev)
    if (rev - 1 not in parents) :
        # one of the parents was the old tip
        return False
    if parents == (nullrev, nullrev):
        return True
    if (len(cl.heads(cl.node(parents[0]))) > 1 and (parents[1] == nullrev or
            len(cl.heads(cl.node(parents[1]))) > 1)):
        return True
    return False

cmdtable = {
    '^workon|task': (doWorkon, WORKON_OPTIONS, 'hg workon TASK_DESCRIPTION'),
    'complete': (doComplete, COMPLETE_OPTIONS, 'hg complete'),
    'synchronise': (doSynchronise, SYNCHRONISE_OPTIONS,
            'hg synchronise [DEST]'),
    'upsync': (doUpsync, UPSYNC_OPTIONS, 'hg upsync [DEST]'),
}
